"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initCtxJsonFromOptions = exports.cli = void 0;
const constants_1 = require("./constants");
const utils_1 = require("./utils");
const evmAtomicTx_1 = require("./evmAtomicTx");
const pvmAtomicTx_1 = require("./pvmAtomicTx");
const addValidator_1 = require("./addValidator");
const addDelegator_1 = require("./addDelegator");
const key_1 = require("./ledger/key");
const sign_1 = require("./ledger/sign");
const forDefi_1 = require("./forDefi");
const withdrawal_1 = require("./withdrawal");
const output_1 = require("./output");
const DERIVATION_PATH = "m/44'/60'/0'/0/0"; // derivation path for ledger
const FLR = 1e9; // one FLR in nanoFLR
const MAX_TRANSCTION_FEE = FLR;
function cli(program) {
    return __awaiter(this, void 0, void 0, function* () {
        // global configurations
        program
            .option("--network <network>", "Network name (flare or costwo)", 'flare')
            .option("--ledger", "Use ledger to sign transactions")
            .option("--blind", "Blind signing (used for ledger)", false)
            .option("--get-hacked", "Use the .env file with the exposed private key")
            .option("--ctx-file <file>", "Context file as returned by init-ctx", 'ctx.json')
            .option("--env-path <path>", "Path to the .env file");
        // context setup
        program
            .command("init-ctx").description("Initialize context file")
            .option("-p, --public-key <public-key>", "Public key of the account")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            initCtxJsonFromOptions(options);
        }));
        // information about the network
        program
            .command("info").description("Relevant information")
            .argument("<type>", "Type of information")
            .action((type) => __awaiter(this, void 0, void 0, function* () {
            const options = getOptions(program, program.opts());
            const ctx = yield contextFromOptions(options);
            if (type == 'addresses') {
                logAddressInfo(ctx);
            }
            else if (type == 'balance') {
                yield logBalanceInfo(ctx);
            }
            else if (type == 'network') {
                logNetworkInfo(ctx);
            }
            else if (type == 'validators') {
                yield logValidatorInfo(ctx);
            }
            else {
                (0, output_1.logError)(`Unknown information type ${type}`);
            }
        }));
        // transaction construction and sending
        program
            .command("transaction").description("Move funds from one chain to another")
            .argument("<type>", "Type of a crosschain transaction")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .option("-a, --amount <amount>", "Amount to transfer")
            .option("-f, --fee <fee>", "Transaction fee")
            .option("-n, --node-id <nodeId>", "The id of the node to stake/delegate to")
            .option("-s, --start-time <start-time>", "Start time of the staking/delegating process")
            .option("-e, --end-time <end-time>", "End time of the staking/delegating process")
            .option("--nonce <nonce>", "Nonce of the constructed transaction")
            .option("--delegation-fee <delegation-fee>", "Delegation fee defined by the deployed validator", "10")
            .option("--threshold <threshold>", "Threshold of the constructed transaction", "1")
            .action((type, options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            if (options.getHacked) {
                // this is more of a concept for future development, by now private key was already exposed to dependencies
                const response = yield (0, utils_1.getUserInput)(`
          You are about to expose your private key to 800+ dependencies, and we cannot guarantee one of them is not malicious!
          This command is not meant to be used in production, but for testing only! Proceed? (Y/N) `);
                if (response == 'Y')
                    yield cliBuildAndSendTxUsingPrivateKey(type, ctx, options);
            }
            else if (options.ledger) {
                yield cliBuildAndSendTxUsingLedger(type, ctx, options, options.blind);
            }
            else {
                yield cliBuildUnsignedTxJson(type, ctx, options.transactionId, options);
            }
        }));
        // signed transaction sending
        program
            .command("send").description("Send signed transaction json to the node")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to send to the network")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            yield cliSendSignedTxJson(ctx, options.transactionId);
        }));
        // forDefi signing
        program
            .command("forDefi").description("Sign with ForDefi")
            .argument("<type>", "Type of a forDefi transaction")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .option("--withdrawal", "Withdrawing funds from c-chain")
            .action((type, options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            if (type == 'sign') {
                if (options.withdrawal) {
                    yield signForDefi(options.transactionId, options.ctxFile, true);
                }
                else {
                    yield signForDefi(options.transactionId, options.ctxFile);
                }
            }
            else if (type == 'fetch') {
                if (options.withdrawal) {
                    yield fetchForDefiTx(options.transactionId, true);
                }
                else {
                    yield fetchForDefiTx(options.transactionId);
                }
            }
        }));
        // withdrawal from c-chain
        program
            .command("withdrawal").description("Withdraw funds from c-chain")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .option("-a, --amount <amount>", "Amount to transfer")
            .option("-t, --to <to>", "Address to send funds to")
            .option("--nonce <nonce>", "Nonce of the constructed transaction")
            .option("--send-signed-tx", "Send signed transaction json to the node")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            if (options.sendSignedTx) {
                yield withdraw_useSignature(ctx, options.transactionId);
            }
            else { // create unsigned transaction
                yield withdraw_getHash(ctx, options.to, options.amount, options.transactionId, options.nonce);
            }
        }));
        // ledger two-step manual signing
        program
            .command("sign-hash").description("Sign a transaction hash (blind signing)")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            yield (0, sign_1.signId)(options.transactionId, DERIVATION_PATH, true);
            (0, output_1.logSuccess)("Transaction signed");
        }));
        program
            .command("sign").description("Sign a transaction (non-blind signing)")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            yield (0, sign_1.signId)(options.transactionId, DERIVATION_PATH, false);
            (0, output_1.logSuccess)("Transaction signed");
        }));
    });
}
exports.cli = cli;
function contextFromOptions(options) {
    return __awaiter(this, void 0, void 0, function* () {
        if (options.ledger) {
            (0, output_1.logInfo)("Fetching account from ledger...");
            const account = yield (0, key_1.ledgerGetAccount)(DERIVATION_PATH, options.network);
            const context = (0, constants_1.getContext)(options.network, account.publicKey);
            return context;
        }
        else if (options.envPath) {
            return (0, constants_1.contextEnv)(options.envPath, options.network);
        }
        else {
            return (0, constants_1.contextFile)(options.ctxFile);
        }
    });
}
function getOptions(program, options) {
    const allOptions = Object.assign(Object.assign({}, program.opts()), options);
    // amount and fee are given in FLR, transform into nanoFLR (FLR = 1e9 nanoFLR)
    if (allOptions.amount) {
        allOptions.amount = (0, utils_1.decimalToInteger)(allOptions.amount.replace(/,/g, ''), 9);
    }
    if (allOptions.fee) {
        allOptions.fee = (0, utils_1.decimalToInteger)(allOptions.fee, 9);
    }
    return allOptions;
}
function capFeeAt(cap, usedFee, specifiedFee) {
    if (usedFee !== specifiedFee) { // if usedFee was that specified by the user, we don't cap it
        const usedFeeNumber = Number(usedFee); // if one of the fees is defined, usedFee is defined
        if (usedFeeNumber > cap)
            throw new Error(`Used fee of ${usedFeeNumber / FLR} FLR is higher than the maximum allowed fee of ${cap / FLR} FLR`);
        (0, output_1.logInfo)(`Using fee of ${usedFeeNumber / FLR} FLR`);
    }
}
//////////////////////////////////////////////////////////////////////////////////////////
// transaction-type translators
function buildUnsignedTxJson(transactionType, context, params) {
    switch (transactionType) {
        case 'exportCP': {
            return (0, evmAtomicTx_1.getUnsignedExportTxCP)(context, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.fee), (params.nonce === undefined) ? undefined : Number(params.nonce));
        }
        case 'importCP':
            return (0, pvmAtomicTx_1.getUnsignedImportTxCP)(context, Number(params.threshold));
        case 'exportPC': {
            return (0, pvmAtomicTx_1.getUnsignedExportTxPC)(context, (0, utils_1.toBN)(params.amount), Number(params.threshold));
        }
        case 'importPC': {
            return (0, evmAtomicTx_1.getUnsignedImportTxPC)(context, (0, utils_1.toBN)(params.fee));
        }
        case 'stake': {
            return (0, addValidator_1.getUnsignedAddValidator)(context, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.delegationFee), Number(params.threshold));
        }
        case 'delegate': {
            return (0, addDelegator_1.getUnsignedAddDelegator)(context, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.threshold));
        }
        default:
            throw new Error(`Unknown transaction type: ${transactionType}`);
    }
}
function sendSignedTxJson(context, signedTxJson) {
    return __awaiter(this, void 0, void 0, function* () {
        switch (signedTxJson.transactionType) {
            case 'exportCP': {
                const { chainTxId } = yield (0, evmAtomicTx_1.issueSignedEvmTxCPExport)(context, signedTxJson);
                return chainTxId;
            }
            case 'importPC': {
                const { chainTxId } = yield (0, evmAtomicTx_1.issueSignedEvmTxPCImport)(context, signedTxJson);
                return chainTxId;
            }
            case 'exportPC':
            case 'importCP':
            case 'stake':
            case 'delegate': {
                const { chainTxId } = yield (0, pvmAtomicTx_1.issueSignedPvmTx)(context, signedTxJson);
                return chainTxId;
            }
            default:
                throw new Error(`Unknown transaction type: ${signedTxJson.transactionType}`);
        }
    });
}
function buildAndSendTxUsingPrivateKey(transactionType, context, params) {
    return __awaiter(this, void 0, void 0, function* () {
        if (transactionType === 'exportCP') {
            return (0, evmAtomicTx_1.exportTxCP)(context, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.fee));
        }
        else if (transactionType === 'importCP') {
            return (0, pvmAtomicTx_1.importTxCP)(context, Number(params.threshold));
        }
        else if (transactionType === 'exportPC') {
            return (0, pvmAtomicTx_1.exportTxPC)(context, (0, utils_1.toBN)(params.amount), Number(params.threshold));
        }
        else if (transactionType === 'importPC') {
            return (0, evmAtomicTx_1.importTxPC)(context, (0, utils_1.toBN)(params.fee));
        }
        else if (transactionType === 'stake') {
            return (0, addValidator_1.addValidator)(context, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.delegationFee), Number(params.threshold));
        }
        else if (transactionType === 'delegate') {
            return (0, addDelegator_1.addDelegator)(context, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.threshold));
        }
        else {
            throw new Error(`Unknown transaction type ${transactionType}`);
        }
    });
}
//////////////////////////////////////////////////////////////////////////////////////////
// initializing ctx.json
function initCtxJsonFromOptions(options) {
    return __awaiter(this, void 0, void 0, function* () {
        let contextFile;
        if (options.ledger) {
            const { publicKey, address } = yield (0, key_1.ledgerGetAccount)(DERIVATION_PATH, options.network);
            const ethAddress = (0, utils_1.publicKeyToEthereumAddressString)(publicKey);
            contextFile = { publicKey, ethAddress, flareAddress: address, network: options.network };
        }
        else if (options.publicKey) {
            if (!(0, utils_1.validatePublicKey)(options.publicKey))
                return (0, output_1.logError)('Invalid public key');
            contextFile = { publicKey: options.publicKey, network: options.network };
        }
        else {
            throw new Error('Either --ledger or --public-key must be specified');
        }
        (0, utils_1.initCtxJson)(contextFile);
        (0, output_1.logSuccess)("Context file created");
    });
}
exports.initCtxJsonFromOptions = initCtxJsonFromOptions;
//////////////////////////////////////////////////////////////////////////////////////////
// Network info
function logAddressInfo(ctx) {
    const [pubX, pubY] = ctx.publicKey;
    const compressedPubKey = (0, utils_1.compressPublicKey)(pubX, pubY).toString('hex');
    (0, output_1.logInfo)(`Addresses on the network "${ctx.config.hrp}"`);
    (0, output_1.log)(`P-chain address: ${ctx.pAddressBech32}`);
    (0, output_1.log)(`C-chain address hex: ${ctx.cAddressHex}`);
    (0, output_1.log)(`secp256k1 public key: 0x${compressedPubKey}`);
}
function logBalanceInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        let cbalance = ((0, utils_1.toBN)(yield ctx.web3.eth.getBalance(ctx.cAddressHex))).toString();
        let pbalance = ((0, utils_1.toBN)((yield ctx.pchain.getBalance(ctx.pAddressBech32)).balance)).toString();
        cbalance = (0, utils_1.integerToDecimal)(cbalance, 18);
        pbalance = (0, utils_1.integerToDecimal)(pbalance, 9);
        (0, output_1.logInfo)(`Balances on the network "${ctx.config.hrp}"`);
        (0, output_1.log)(`C-chain ${ctx.cAddressHex}: ${cbalance}`);
        (0, output_1.log)(`P-chain ${ctx.pAddressBech32}: ${pbalance}`);
    });
}
function logNetworkInfo(ctx) {
    const pchainId = ctx.pchain.getBlockchainID();
    const cchainId = ctx.cchain.getBlockchainID();
    (0, output_1.logInfo)(`Information about the network "${ctx.config.hrp}"`);
    (0, output_1.log)(`blockchainId for P-chain: ${pchainId}`);
    (0, output_1.log)(`blockchainId for C-chain: ${cchainId}`);
    (0, output_1.log)(`assetId: ${ctx.avaxAssetID}`);
}
function logValidatorInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const pending = yield ctx.pchain.getPendingValidators();
        const current = yield ctx.pchain.getCurrentValidators();
        const fpending = JSON.stringify(pending, null, 2);
        const fcurrent = JSON.stringify(current, null, 2);
        (0, output_1.logInfo)(`Validators on the network "${ctx.config.hrp}"`);
        (0, output_1.log)(`pending: ${fpending}`);
        (0, output_1.log)(`current: ${fcurrent}`);
    });
}
//////////////////////////////////////////////////////////////////////////////////////////
// Transaction building and execution
function cliBuildAndSendTxUsingLedger(transactionType, context, params, blind) {
    return __awaiter(this, void 0, void 0, function* () {
        (0, output_1.logInfo)("Creating export transaction...");
        const unsignedTxJson = yield buildUnsignedTxJson(transactionType, context, params);
        capFeeAt(MAX_TRANSCTION_FEE, unsignedTxJson.usedFee, params.fee);
        (0, output_1.logInfo)("Please review and sign the transaction on your ledger device...");
        const { signature } = yield (0, sign_1.ledgerSign)(unsignedTxJson, DERIVATION_PATH, blind);
        const signedTxJson = Object.assign(Object.assign({}, unsignedTxJson), { signature });
        (0, output_1.logInfo)("Sending transaction to the node...");
        const chainTxId = yield sendSignedTxJson(context, signedTxJson);
        (0, output_1.logSuccess)(`Transaction with id ${chainTxId} sent to the node`);
    });
}
function cliBuildUnsignedTxJson(transactionType, ctx, id, params) {
    return __awaiter(this, void 0, void 0, function* () {
        const unsignedTxJson = yield buildUnsignedTxJson(transactionType, ctx, params);
        capFeeAt(MAX_TRANSCTION_FEE, unsignedTxJson.usedFee, params.fee);
        (0, utils_1.saveUnsignedTxJson)(unsignedTxJson, id);
        (0, output_1.logSuccess)(`Unsigned transaction with id ${id} constructed`);
    });
}
function cliSendSignedTxJson(ctx, id) {
    return __awaiter(this, void 0, void 0, function* () {
        const chainTxId = yield sendSignedTxJson(ctx, (0, utils_1.readSignedTxJson)(id));
        (0, output_1.logSuccess)(`Signed transaction ${id} with id ${chainTxId} sent to the node`);
    });
}
function cliBuildAndSendTxUsingPrivateKey(transactionType, ctx, params) {
    return __awaiter(this, void 0, void 0, function* () {
        const { txid, usedFee } = yield buildAndSendTxUsingPrivateKey(transactionType, ctx, params);
        if (usedFee)
            (0, output_1.logInfo)(`Used fee of ${(0, utils_1.integerToDecimal)(usedFee, 9)} FLR`);
        (0, output_1.logSuccess)(`Transaction with id ${txid} built and sent to the network`);
    });
}
//////////////////////////////////////////////////////////////////////////////////////////
// Transaction execution using ForDefi api
function signForDefi(transaction, ctx, withdrawal = false) {
    return __awaiter(this, void 0, void 0, function* () {
        const txid = yield (0, forDefi_1.sendToForDefi)(transaction, ctx, withdrawal);
        (0, output_1.logSuccess)(`Transaction with id ${txid} sent to the node`);
    });
}
function fetchForDefiTx(transaction, withdrawal = false) {
    return __awaiter(this, void 0, void 0, function* () {
        const signature = yield (0, forDefi_1.getSignature)(transaction, withdrawal);
        (0, output_1.logSuccess)(`Success! Signature: ${signature}`);
    });
}
function withdraw_getHash(ctx, to, amount, id, nonce) {
    return __awaiter(this, void 0, void 0, function* () {
        const fileId = yield (0, withdrawal_1.createWithdrawalTransaction)(ctx, to, amount, id, nonce);
        (0, output_1.logSuccess)(`Transaction with id ${fileId} constructed`);
    });
}
function withdraw_useSignature(ctx, id) {
    return __awaiter(this, void 0, void 0, function* () {
        const txId = yield (0, withdrawal_1.sendSignedWithdrawalTransaction)(ctx, id);
        (0, output_1.logSuccess)(`Transaction with id ${txId} sent to the node`);
    });
}
